from __future__ import annotations

import os
import json
import datetime
from typing import Union, List, Dict, Any, Optional, Callable

import psycopg2
import psycopg2.extensions
import psycopg2.extras
import psycopg2.errors

try:
    import dotenv
    dotenv.load_dotenv('.env')
except ModuleNotFoundError:
    pass


def get_postgres_from_env() -> Postgres:
    """
    Returns a Postgres object with credentials from environment variables.

    If environment variables are not set, it uses default values.

    Returns:
        Postgres: A Postgres object with connection details.
    """
    return Postgres(
        host=os.environ.get('POSTGRES_HOST', 'localhost'),
        port=os.environ.get('POSTGRES_PORT', '5432'),
        user=os.environ.get('POSTGRES_USER', 'daniel'),
        password=os.environ.get('POSTGRES_PASSWORD', 'MyPassword123'),
        database=os.environ.get('POSTGRES_DATABASE', 'my_db')
    )


def sort(val):
    """
    Assigns a sorting priority to PostgreSQL data types.

    Args:
        val (str): A PostgreSQL data type.

    Returns:
        int: The sorting priority of the data type.
    """
    _sort = {'text': 0, 'real': 1, 'bigint': 2, 'integer': 3, 'json': 4, 'timestamp': 5, 'date': 6, 'boolean': 7,
             'character(10)': 8, 'character': 8}
    return _sort[val] if val in _sort else 0


def get_type(val):
    """
    Determines the PostgreSQL data type for a given Python value.

    Args:
        val: A Python value of any type.

    Returns:
        str: The corresponding PostgreSQL data type.
    """
    if val is None:
        return 'boolean'  # 'character(10)'
    if isinstance(val, bool):
        return 'boolean'
    if isinstance(val, str):
        return 'text'
    if isinstance(val, float):
        return 'real'
    if isinstance(val, int):
        if -9223372036854775805 > val or val > 9223372036854775805:
            return 'text'
        if -2147483645 > val or val > 2147483645:
            return 'bigint'
        return 'integer'
    if isinstance(val, (list, dict)):
        return 'json'
    if isinstance(val, datetime.datetime):
        return 'timestamptz'  # 'timestamp'
    if isinstance(val, datetime.date):
        return 'date'
    if isinstance(val, bytes):
        return 'bytea'
    return 'text'


def guess_data_type(column: List):
    """
    Guesses the most appropriate data type for a column of values.

    Args:
        column (List): A list of values from a single column.

    Returns:
        str: The guessed PostgreSQL data type for the column.
    """
    v = {get_type(cell) for cell in column}
    v = list(sorted(v, key=lambda a: sort(a)))
    v = v[0]
    return v


def guess_table_schema(table: List[Dict]):
    """
    Guesses the schema for a table represented as a list of dictionaries.

    Args:
        table (List[Dict]): A list of dictionaries representing table rows.

    Returns:
        Dict[str, str]: A dictionary mapping column names to their guessed data types.
    """
    vv = {}
    for row in table:
        for k, v in row.items():
            if k not in vv:
                vv[k] = []
            vv[k].append(v)
    return {k: guess_data_type(v) for k, v in vv.items()}


class Postgres:
    """
    A class to handle PostgreSQL database operations.

    Attributes:
        host (str): The database host.
        port (str): The database port.
        user (str): The database user.
        password (str): The database password.
        database (str): The database name.
    """
    host: str = 'localhost'
    port: str = '5432'
    user: str = 'daniel'
    password: str = 'MyPassword123'
    database: str = 'my_db'

    def __init__(self, host=None, port=None, user=None, password=None, database=None):
        """
        Initializes the Postgres object with connection details.

        Args:
            host (str, optional): The database host.
            port (str, optional): The database port.
            user (str, optional): The database user.
            password (str, optional): The database password.
            database (str, optional): The database name.
        """
        self.host = host or self.host
        self.port = port or self.port
        self.user = user or self.user
        self.password = password or self.password
        self.database = database or self.database

    def connect(self):
        return psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database
        )

    def query(self, query: str, *args):
        """
        Executes a SQL query.

        Args:
            query (str): The SQL query to execute.
            conn (connection, optional): An existing database connection to use.
            cur (cursor, optional): An existing database cursor to use.

        Returns:
            List[Tuple]: The result of the query.
        """
        with psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database
        ) as conn:
            cur = conn.cursor()

            cur.execute(query, args)
            r = []
            try:
                r = cur.fetchall()
            except psycopg2.ProgrammingError: pass
            conn.commit()

            return r

    def download_table(self, table_name: str = None, columns='*', suffix='', sql=None):
        """
        Downloads data from a PostgreSQL table.

        Args:
            table_name (str, optional): The name of the table to download.
            columns (str, optional): The columns to select. Defaults to '*'.
            suffix (str, optional): Additional SQL to append to the query.
            sql (str, optional): A custom SQL query to execute instead of selecting from a table.

        Returns:
            List[Dict]: A list of dictionaries representing the table rows.
        """
        query = (sql or f'select {columns} from {table_name} {suffix};')

        with psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database
        ) as conn:
            cur = conn.cursor()

            cur.execute(query)

            table = cur.fetchall()
            column_names = tuple(col[0] for col in cur.description)

            def check_values(values):
                def check_value(value):
                    if isinstance(value, str):
                        value = value.replace("''", "'")
                    return value
                return [check_value(value) for value in values]

            return [dict(zip(column_names, check_values(row))) for row in table]

    def upload_table(
            self,
            table_name: str,
            table: List[Dict],
            partition: str | None = None,
            partition_type: str | None = 'LIST',
            partition_query: str | None = None,
            id_column=None
    ) -> None:
        """
        Uploads data to a PostgreSQL table.

        Args:
            table_name (str): The name of the table to upload to.
            table (List[Dict]): The data to upload, as a list of dictionaries.
            partition (str, optional): The column to use for partitioning.
            partition_type (str, optional): The type of partitioning to use. Defaults to 'LIST'.
            partition_query (str, optional): A custom SQL query to execute for partitioning.
                EXAMPLE: "CREATE TABLE IF NOT EXISTS "table_name_val" PARTITION OF "table_name" FOR VALUES IN ('val');"
            id_column (str, optional): The name of the ID column.
        """
        def convert_value(value):
            if isinstance(value, str):
                return value.replace("'", "''")
            if isinstance(value, (int, float, bool, type(None), datetime.datetime)):
                return value
            elif isinstance(value, (tuple, list, dict)):
                return f"{json.dumps(value)}"
            elif isinstance(value, (datetime.datetime, datetime.date)):
                return value
            elif isinstance(value, bytes):
                return value
            raise ValueError(f'cannot place {value} in type "{type(value)}"')
        # check for multiple id_columns
        ids = tuple(set(id_column)) if isinstance(id_column, (set, tuple, list)) else (id_column, )
        id_column = '", "'.join(set(id_column)) if isinstance(id_column, (set, tuple, list)) else id_column

        # get all headers
        keys = {}
        for row in table:
            for key in row:
                if key not in keys:
                    keys[key] = None

        # put headers in order, giving None values to missing headers
        for i, row in enumerate(table):
            table[i] = {k: convert_value(row.get(k, None)) for k in keys}

        with psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database
        ) as conn:
            cur = conn.cursor()
            self.check_for_table(conn, cur, table_name, table, id_column, partition, partition_type)

            if partition_query:
                try:
                    cur.execute(partition_query)
                except psycopg2.errors.DuplicateTable:
                    for _query in partition_query.split(';'):
                        try:
                            cur.execute(_query)
                        except psycopg2.errors.DuplicateTable:
                            pass
                conn.commit()
            keys = ', '.join([f'"{k}"' for k in table[0].keys()])
            vals = ', '.join(['%s'] * len(table[0].keys()))
            q = f'INSERT INTO {table_name} ({keys}) VALUES ({vals})'
            if id_column is not None:
                f = {*ids}
                if partition:
                    f.add(partition)
                def _filter(k):
                    return k not in f  # {id_column, partition}  # k != id_column
                conflict = f'("{id_column}")' if not partition or partition == id_column else f'("{id_column}", "{partition}")'
                pers = ['%s'] * (len(tuple(filter(_filter, table[0].keys()))))  # - (1 + (not not partition)))
                p1, p2 = '(' if len(pers) > 1 else '', ')' if len(pers) > 1 else ''
                q = f'''INSERT INTO {table_name} ({keys}) VALUES ({vals})
            ON CONFLICT {conflict} DO UPDATE SET {p1}{', '.join([f'"{k}"' for k in filter(_filter, table[0].keys())])}{p2} = {p1}{', '.join(pers)}{p2};'''
                table = (tuple(list(row.values()) + [v for k, v in row.items() if _filter(k)  # k != id_column
                                                     ]) for row in table)

            else:
                table = (tuple(row.values()) for row in table)
            # table = (tuple(row.values()) for row in table)

            psycopg2.extras.execute_batch(cur, q, table)

            conn.commit()

    def table_schema(self, table_name: str, cur):
        """
        Retrieves the schema of a table.

        Args:
            table_name (str): The name of the table.
            cur: A database cursor.

        Returns:
            List[Tuple]: A list of tuples containing column names and data types.
        """
        query = f"SELECT column_name, data_type FROM information_schema.columns where table_name = '{table_name}' ORDER BY ordinal_position;"
        cur.execute(query)
        return cur.fetchall()  # tuple(map(lambda row: row[0], cur.fetchall()))

    def check_for_table(self, conn, cur, table_name: str, table: List[Dict], id_column=None, partition: str = None,
                        partition_type: str = 'LIST', skip_alterations=False):
        """
        Checks if a table exists and creates or alters it as necessary.

        Args:
            conn: A database connection.
            cur: A database cursor.
            table_name (str): The name of the table to check.
            table (List[Dict]): The data to be inserted into the table.
            id_column (str, optional): The name of the ID column.
            partition (str, optional): The column to use for partitioning.
            partition_type (str, optional): The type of partitioning to use. Defaults to 'LIST'.
            skip_alterations (bool, optional): If True, skips table alterations. Defaults to False.
        """
        schema = guess_table_schema(table)

        cur.execute(
            f"SELECT EXISTS (SELECT 1 FROM pg_tables  WHERE tablename = '{table_name}') AS table_existence;")
        table_exists = cur.fetchall()[0][0]

        if not table_exists:
            try:
                primary = ''
                unique = ''
                if partition:
                    if id_column:
                        if id_column == partition or partition in [s.strip() for s in id_column.replace('"', '').split(',')]:
                            unique = f', PRIMARY KEY ("{id_column}")'
                        else:
                            unique = f', PRIMARY KEY ("{id_column}", "{partition}")'
                else:
                    primary = ' primary key UNIQUE'
                    if id_column:
                        unique = f', UNIQUE ("{id_column}")'

                end = ';' if not isinstance(partition, str) else f' PARTITION BY {partition_type} ("{partition}");'

                def column_and_data_type(k, v):
                    if k == id_column and not partition:
                        return f'"{k}" {v}{primary}'
                    return f'"{k}" {v}'

                sql = f'''create table if not exists {table_name}({", ".join([column_and_data_type(k, v) for k, v in schema.items()])}{unique}){end}'''

                cur.execute(sql)
                conn.commit()
            except psycopg2.errors.UniqueViolation as e:
                print(e)
        elif not skip_alterations:
            db_schema = dict(self.table_schema(table_name, cur))
            conn.commit()

            for col in set(schema.keys()) ^ set(db_schema.keys()):
                if col in schema:
                    try:
                        cur.execute(f'ALTER TABLE {table_name} ADD COLUMN "{col}" {schema[col]};')
                        conn.commit()
                    except psycopg2.errors.DuplicateColumn as e:
                        print(e)
                elif col in db_schema:
                    pass

            for col in set(schema.keys()) & set(db_schema.keys()):
                if schema[col] != db_schema[col]:
                    if sort(schema[col]) < sort(db_schema[col]):
                        # print(f'comparing {sort(schema[col])} < {sort(db_schema[col])} == sort({schema[col]}) < sort({db_schema[col]})')
                        self.convert_table_column_type(table_name, f'"{col}"', schema[col], conn, cur)
                    elif sort(schema[col]) > sort(db_schema[col]):
                        pass

    def convert_table_column_type(self, table_name: str, column_name: str, new_type: str, conn, cur):
        """
        Converts the data type of a table column.

        Args:
            table_name (str): The name of the table.
            column_name (str): The name of the column to convert.
            new_type (str): The new data type for the column.
            conn: A database connection.
            cur: A database cursor.
        """
        base_values = {'text': "''", 'real': '0', 'bigint': '0', 'integer': '0', 'json': "''",
                     'timestamp': datetime.datetime.utcnow(), 'date': "'2023-01-01'", 'boolean': 'false', 'character(10)': "''"}
        conn.commit()
        try:
            cur.execute(f'ALTER TABLE {table_name} ALTER COLUMN {column_name} TYPE {new_type};')
            conn.commit()
        except psycopg2.errors.DatatypeMismatch as e:
            base_value = base_values[new_type]
            try:
                conn.rollback()
                cur.execute(f'ALTER TABLE {table_name} '
                            f'ALTER COLUMN {column_name} '
                            f'TYPE {new_type} '
                            f'USING (nullif({column_name}, {base_value}))::{new_type};')
                conn.commit()
            except:  # (psycopg2.errors.InFailedSqlTransaction, psycopg2.errors.CannotCoerce) as e:
                conn.rollback()
                cur.execute(f'ALTER TABLE {table_name} '
                            f'ALTER COLUMN {column_name} '
                            f'TYPE {new_type} '
                            f'USING {column_name}::text::{new_type};')
                conn.commit()




"""
This module implements a socket server for sending and receiving byte data using keys.

The server allows clients to set, get, and delete data associated with specific keys.
Data is stored in files within a '.bucket' directory.
"""

import socket
import threading
import os
import sys
import time

import psutil
import redis
import psycopg2
from kazoo.client import KazooClient
import kazoo.exceptions
import tqdm

import buelon.helpers.postgres

try:
    import dotenv
    dotenv.load_dotenv('.env')
except ModuleNotFoundError:
    pass


# Environment variables for client and server configuration
USING_POSTGRES: bool = os.environ.get('USING_POSTGRES_BUCKET', 'false') == 'true'
POSTGRES_TABLE: str = os.environ.get('POSTGRES_TABLE', 'buelon_bucket')

REDIS_HOST: str = os.environ.get('REDIS_HOST', 'null')
REDIS_PORT: int = int(os.environ.get('REDIS_PORT', 6379))
REDIS_DB: int = int(os.environ.get('REDIS_DB', 0))
REDIS_EXPIRATION: int | None = os.environ.get('REDIS_EXPIRATION', 60*60*24*7)
try:
    REDIS_EXPIRATION = int(REDIS_EXPIRATION)
except ValueError:
    REDIS_EXPIRATION = None
USING_REDIS: bool = os.environ.get('USING_REDIS', 'false') == 'true'  # False  # REDIS_HOST != 'null'

BUCKET_CLIENT_HOST: str = os.environ.get('BUCKET_CLIENT_HOST', 'localhost')
BUCKET_CLIENT_PORT: int = int(os.environ.get('BUCKET_CLIENT_PORT', 61535))

BUCKET_SERVER_HOST: str = os.environ.get('BUCKET_SERVER_HOST', '0.0.0.0')
BUCKET_SERVER_PORT: int = int(os.environ.get('BUCKET_SERVER_PORT', 61535))

USING_ZOOKEEPER = os.environ.get('USING_ZOOKEEPER', 'false') == 'true'
ZOOKEEPER_HOSTS: str = os.environ.get('ZOOKEEPER_HOSTS', 'localhost:2181')
ZOOKEEPER_PATH: str = f"{os.environ.get('ZOOKEEPER_PATH', '/buelon/bucket')}"

PERSISTENT_PATH: str = f"{os.environ.get('PERSISTENT_PATH', '__PERSISTENT__')}"

BUCKET_END_TOKEN = b'[-_-]'
BUCKET_SPLIT_TOKEN = b'[*BUCKET_SPLIT_TOKEN*]'

save_path = os.path.join('.bue', 'bucket')

database: dict[str, bytes] = {}
database_keys_in_order = []
# MAX_DATABASE_SIZE: int = min(1024 * 1024 * 1024 * 1, int(psutil.virtual_memory().total / 8))
MAX_DATABASE_SIZE: int = 50 * 1024 * 1024

if not USING_REDIS and not USING_POSTGRES and not USING_ZOOKEEPER:
    if not os.path.exists(save_path):
        os.makedirs(save_path)


def receive(conn: socket.socket, size: int = 1024) -> bytes:
    """
    Receive data from a socket connection until the end marker is found.

    Args:
        conn (socket.socket): The socket connection to receive data from.
        size (int, optional): The maximum number of bytes to receive at once. Defaults to 1024.

    Returns:
        bytes: The received data without the end marker.
    """
    data = b''
    while not data.endswith(BUCKET_END_TOKEN):
        v = conn.recv(size)
        data += v
    token_len = len(BUCKET_END_TOKEN)
    return data[:-token_len]


def send(conn: socket.socket, data: bytes) -> None:
    """
    Send data through a socket connection with an end marker.

    Args:
        conn (socket.socket): The socket connection to send data through.
        data (bytes): The data to be sent.
    """
    conn.sendall(data + BUCKET_END_TOKEN)


def retry_connection(func: callable):

    """
    Decorator to retry a function call if a ConnectionResetError occurs.

    Args:
        func (callable): The function to be decorated.

    Returns:
        callable: The decorated function.
    """
    def wrapper(self, *args, **kwargs) -> bytes | None:
        tries = 4
        kwargs['timeout'] = kwargs.get('timeout', 60 * 5.)
        for i in range(tries):
            try:
                return func(self, *args, **kwargs)
            except (TimeoutError, ConnectionResetError):
                kwargs['timeout'] *= 2
            except psycopg2.OperationalError:
                time.sleep((i + 1) * 5.)
                self.db = buelon.helpers.postgres.get_postgres_from_env()
        raise
    return wrapper


def redis_secure_connection(func: callable):
    def wrapper(*args, **kwargs) -> bytes | None:
        try:
            return func(*args, **kwargs)
        except redis.exceptions.ConnectionError:
            self: Client = args[0]
            try:
                del self.redis_client
            except Exception as e:
                print(f'Change `Exception` to {type(e)}. e: {e}')
            self.redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=0)
            return func(*args, **kwargs)
    return wrapper


class Client:
    """A client for interacting with the bucket server."""
    PORT: int
    HOST: str

    db: buelon.helpers.postgres.Postgres

    def __init__(self):
        """Initialize the client with host and port from environment variables."""
        self.PORT = BUCKET_CLIENT_PORT
        self.HOST = BUCKET_CLIENT_HOST

        if USING_ZOOKEEPER:
            self.zk = KazooClient(hosts=ZOOKEEPER_HOSTS)
            self.zk.start(timeout=60 * 15)
            self.zk.ensure_path(ZOOKEEPER_PATH)
        elif USING_POSTGRES:
            self.db = buelon.helpers.postgres.get_postgres_from_env()
            try:
                with self.db.connect() as conn:
                    cur = conn.cursor()
                    cur.execute(f'CREATE TABLE IF NOT EXISTS {POSTGRES_TABLE} (key TEXT PRIMARY KEY, data BYTEA, epoch REAL);')
                    if '.' in POSTGRES_TABLE:
                        args = POSTGRES_TABLE.split('.')
                        schema = args[-2]
                        table_name = args[-1]
                    else:
                        schema = 'public'
                        table_name = POSTGRES_TABLE
                    # cur.execute(f'CREATE INDEX IF NOT EXISTS idx_{POSTGRES_TABLE}_key ON {POSTGRES_TABLE} USING hash (key);')
                    cur.execute(f'CREATE INDEX IF NOT EXISTS idx_{table_name} ON {POSTGRES_TABLE} USING hash (key);')
                    conn.commit()
            except psycopg2.errors.InsufficientPrivilege:
                print('Insufficient privileges to use postgres bucket.')
        elif USING_REDIS:
            self.redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=0)

    def __getattr__(self, item):
        if item == 'db':
            if USING_POSTGRES:
                if not hasattr(self, 'db'):
                    v = buelon.helpers.postgres.get_postgres_from_env()
                    self.db = v
                    return v

    def __enter__(self):
        return self

    def __del__(self):
        if USING_ZOOKEEPER:
            try:
                self.zk.stop()
                self.zk.close()
            except kazoo.exceptions.WriterNotClosedException:
                pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if USING_ZOOKEEPER:
            try:
                self.zk.stop()
                self.zk.close()
            except kazoo.exceptions.WriterNotClosedException:
                pass
        elif USING_REDIS:
            self.redis_client.connection.disconnect()

    @retry_connection
    def set(self, key: str, data: bytes, timeout: float | None = 60 * 5., persistent: bool = False) -> None:
        """
        Set data for a given key on the server.

        Args:
            key (str): The key to associate with the data.
            data (bytes): The data to store.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.
        """
        if persistent:
            key = PERSISTENT_PATH + '/' + key  # f'{PERSISTENT_PATH}/{key}'
        if USING_ZOOKEEPER:
            if not persistent:
                key = 'temp/' + key  # f'temp/{key}'
            key = ZOOKEEPER_PATH + '/' + key  # f'{ZOOKEEPER_PATH}/{key}'
            self.zk.ensure_path(key)
            self.zk.set(key, data)
        elif USING_POSTGRES:
            with self.db.connect() as conn:
                cur = conn.cursor()
                cur.execute(f'INSERT INTO {POSTGRES_TABLE} (key, data, epoch) VALUES (%s, %s, %s) '
                             'ON CONFLICT (key) DO UPDATE SET (data, epoch) = (EXCLUDED.data, EXCLUDED.epoch)', (key, data, time.time()))
                conn.commit()
        elif USING_REDIS:
            self.redis_set(key, data, True)
        else:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                s.settimeout(timeout)
                s.connect((self.HOST, self.PORT))
                if len(data) < 2048:
                    send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'set', f'{timeout}'.encode(), data]))
                    receive(s)
                else:
                    send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'big-set', f'{timeout}'.encode(), f'{len(data)}'.encode()]))
                    receive(s)
                    send(s, data)
                    receive(s)

    @retry_connection
    def get(self, key: str, timeout: float | None = 60 * 5., persistent: bool = False) -> bytes | None:
        """
        Retrieve data for a given key from the server.

        Args:
            key (str): The key to retrieve data for.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.

        Returns:
            bytes or None: The retrieved data, or None if the key doesn't exist.
        """
        if persistent:
            key = PERSISTENT_PATH + '/' + key  # f'{PERSISTENT_PATH}/{key}'
        if USING_ZOOKEEPER:
            if not persistent:
                key = 'temp/' + key  # f'temp/{key}'
            key = ZOOKEEPER_PATH + '/' + key  # f'{ZOOKEEPER_PATH}/{key}'
            try:
                return self.zk.get(key)[0]
            except (KeyError, kazoo.exceptions.NoNodeError):
                return None
        elif USING_POSTGRES:
            with self.db.connect() as conn:
                cur = conn.cursor()
                cur.execute(f'SELECT data FROM {POSTGRES_TABLE} WHERE key = %s', (key,))
                data = cur.fetchone()
                if not data:
                    return None
                return bytes(data[0])
        if USING_REDIS:
            return self.redis_get(key)
        # raise RuntimeError('Not using redis.')
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.settimeout(timeout)
            s.connect((self.HOST, self.PORT))

            send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'get', f'{timeout}'.encode(), b'__null__']))

            data: bytes = receive(s)
            if data[:len(b'__big__')] == b'__big__':
                size = int(data[len(b'__big__'):])
                send(s, b'ok')
                data = receive(s, size)
                return data
            if data == b'__null__':
                return None
            return data

    @retry_connection
    def delete(self, key: str, timeout: float | None = 60 * 5., persistent: bool = False) -> None:
        """
        Delete data for a given key on the server.

        Args:
            key (str): The key to delete data for.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.
        """
        if persistent:
            key = PERSISTENT_PATH + '/' + key  # f'{PERSISTENT_PATH}/{key}'
        if USING_ZOOKEEPER:
            if not persistent:
                key = 'temp/' + key  # f'temp/{key}'
            key = ZOOKEEPER_PATH + '/' + key  # f'{ZOOKEEPER_PATH}/{key}'
            self.zk.delete(key)
        elif USING_POSTGRES:
            with self.db.connect() as conn:
                cur = conn.cursor()
                cur.execute(f'DELETE FROM {POSTGRES_TABLE} WHERE key = %s', (key, ))
                conn.commit()
            return
        if USING_REDIS:
            self.redis_delete(key)
            return

        # raise RuntimeError('Not using redis.')
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.settimeout(timeout)
            s.connect((self.HOST, self.PORT))

            send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'delete', f'{timeout}'.encode(), b'__null__']))
            receive(s)

    def bulk_set(self, keys_values: dict, save: bool = False, persistent: bool = False):
        """
        Set multiple key-value pairs on the server.

        Args:
            keys_values (dict): A dictionary of key-value pairs to set.
            save (bool, optional): Whether to save the data to disk. Defaults to False.
        """
        if USING_ZOOKEEPER:
            for k, v in tqdm.tqdm(keys_values.items(), desc=f'setting {len(keys_values)} paths'):
                self.set(k, v, persistent=persistent)
            # for path in tqdm.tqdm(keys_values, desc=f'ensuring {len(keys_values)} paths'):
            #     self.zk.ensure_path(path)
            # transaction = self.zk.transaction()
            # for k, v in keys_values.items():
            #     transaction.set_data(k, v)
            #     # transaction.create(k, v, makepath=True)
            # results = transaction.commit()
            # print('done uploading paths')
            # # print('results', results)
        elif USING_POSTGRES:
            return self.postgres_bulk_set(keys_values, save)
        elif USING_REDIS:
            return self.redis_bulk_set(keys_values, save)
        else:
            for k, v in keys_values.items():
                self.set(k, v, save=save, persistent=persistent)

    def clean_up(self):
        """
        Clean up the database by removing expired keys.
        """
        t = time.time()
        object_lifetime = 60 * 60 * 24
        if USING_ZOOKEEPER:
            self.zk.delete(ZOOKEEPER_PATH + '/temp', recursive=True)  # (f'{ZOOKEEPER_PATH}/temp', recursive=True)
            self.zk.ensure_path(ZOOKEEPER_PATH + '/temp')  # (f'{ZOOKEEPER_PATH}/temp')
            base_path = ZOOKEEPER_PATH + '/temp'  # f'{ZOOKEEPER_PATH}/temp'
            for child in self.zk.get_children(base_path):
                path = ZOOKEEPER_PATH + '/temp/' + child  # f'{ZOOKEEPER_PATH}/temp/{child}'
                data, config = self.zk.get(path)
                if (t - (config.mtime / 1000)) > object_lifetime:
                    self.zk.delete(path)
        elif USING_POSTGRES:
            with self.db.connect() as conn:
                cur = conn.cursor()
                cur.execute(f'DELETE FROM {POSTGRES_TABLE} WHERE epoch < {t - object_lifetime} and key not like \'{PERSISTENT_PATH}%\';')
                conn.commit()
        elif USING_REDIS:
            pass
        # else:
        #     for root, dirs, files in os.walk(save_path, topdown=True):
        #         dirs[:] = [d for d in dirs if d != PERSISTENT_PATH]
        #         for file in files:
        #             path = os.path.join(root, file)
        #             if (t - os.path.getmtime(path)) > object_lifetime:
        #                 os.remove(path)

    def postgres_bulk_set(self, keys_values: dict, save: bool = False):
        """
        Set multiple key-value pairs in the PostgreSQL database.

        Args:
            keys_values (dict): A dictionary of key-value pairs to set.
            save (bool, optional): Whether to save the data to disk. Defaults to False.
        """
        with self.db.connect() as conn:
            cur = conn.cursor()
            # psycopg2.extras.execute_batch(cur, q, table)
            q = f'''INSERT INTO {POSTGRES_TABLE} (key, data, epoch) VALUES (%s, %s, %s) 
                ON CONFLICT (key) DO UPDATE SET (data, epoch) = (EXCLUDED.data, EXCLUDED.epoch)'''
            table = ((k, v, time.time()) for k, v in keys_values.items())
            psycopg2.extras.execute_batch(cur, q, table)
            # for k, v in keys_values.items():
            #     cur.execute(f'INSERT INTO {POSTGRES_TABLE} (key,data, epoch) VALUES (%s, %s, %s) '
            #                  'ON CONFLICT (key) DO UPDATE SET (data, epoch) = (EXCLUDED.data, EXCLUDED.epoch)', (k, v, time.time()))
            conn.commit()

    @redis_secure_connection
    def redis_bulk_set(self, keys_values: dict, save: bool = False):
        """
        Set multiple key-value pairs in Redis.

        Args:
            keys_values (dict): A dictionary of key-value pairs to set.
            save (bool, optional): Whether to save the data to disk. Defaults to False.
        """
        self.redis_client.mset(keys_values)

        if save:
            self.redis_client.save()

        for k in keys_values:
            self.redis_client.expire(k, 60*60*24*7)

    @redis_secure_connection
    def redis_set(self, key: str, data: bytes, save: bool = False, expiration: int | None = None) -> None:
        """
        Set data for a given key in Redis.

        Args:
            key (str): The key to associate with the data.
            data (bytes): The data to store.
            save (bool, optional): Whether to save the data to disk. Defaults to False.
            expiration (int, optional): The expiration time for the key in seconds. Negative value will ensure no expiration. Defaults to either environment variable `REDIS_EXPIRATION` or 60 * 60 * 24 * 7 (1 week).
        """
        if not isinstance(expiration, int):
            expiration = REDIS_EXPIRATION
        else:
            if expiration < 0:
                expiration = None

        self.redis_client.set(key, data, ex=expiration)

        if save:
            self.redis_client.save()

    @redis_secure_connection
    def redis_get(self, key: str) -> bytes | None:
        """
        Retrieve data for a given key from Redis.

        Args:
            key (str): The key to retrieve data for.

        Returns:
            bytes or None: The retrieved data, or None if the key doesn't exist.
        """
        data = self.redis_client.get(key)
        if data is None:
            return None
        return data

    @redis_secure_connection
    def redis_delete(self, key: str) -> None:
        """
        Delete data for a given key from Redis.

        Args:
            key (str): The key to delete data for.
        """
        # self.redis_client.delete(key)
        self.redis_client.delete(key)


def check_file_directory(path: str) -> None:
    """
    Ensure the directory for a file path exists, creating it if necessary.

    Args:
        path (str): The file path to check.
    """
    if os.path.dirname(path) == '' or os.path.dirname(path) == '.' or os.path.dirname(path) == '.\\' or os.path.dirname(path) == '/' or os.path.dirname(path) == '\\' or os.path.dirname(path) == './':  # os.path.dirname(path) in {'', '.', '/', '\\', './', '.\\'}:
        return
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))


def server_get(key: str) -> bytes | None:
    """
    Retrieve data for a given key from the server.

    Args:
        key (str): The key to retrieve data for.

    Returns:
        bytes or None: The retrieved data, or None if the key doesn't exist.
    """
    global database, save_path
    if key not in database:
        if not os.path.exists(os.path.join(save_path, key)):
            return b'__null__'
        with open(os.path.join(save_path, key), 'rb') as f:
            database[key] = f.read()
    return database[key]


def server_set(key: str, data: bytes) -> None:
    """
    Set data for a given key on the server.

    Args:
        key (str): The key to associate with the data.
        data (bytes): The data to store.
    """
    global database, save_path
    database[key] = data
    # # This is set later
    # check_file_directory(os.path.join(save_path, key))
    # with open(os.path.join(save_path, key), 'wb') as f:
    #     f.write(data)


def server_delete(key: str) -> None:
    """
    Delete data for a given key from the server.

    Args:
        key (str): The key to delete data for.
    """
    global database, save_path
    if key in database:
        del database[key]
    try:
        os.remove(os.path.join(save_path, key))
    except FileNotFoundError:
        pass


def handle_client(conn: socket.socket) -> None:
    """
    Handle a client connection, processing set, get, and delete requests.

    Args:
        conn (socket.socket): The client connection socket.
    """
    global database, save_path
    k: bytes
    m: bytes
    data: bytes
    key: str
    method: str

    def ok():
        send(conn, b'ok')

    with conn:

        k, m, t, data = receive(conn).split(BUCKET_SPLIT_TOKEN)
        key = k.decode('utf-8')
        method = m.decode('utf-8')
        timeout = float(t)

        conn.settimeout(timeout)

        if method == 'big-set':
            size = int(data)
            ok()
            data = receive(conn, size)
            server_set(key, data)
            ok()
        if method == 'set':
            # # ok()
            # # database[key] = data
            # check_file_directory(os.path.join(save_path, key))
            # with open(os.path.join(save_path, key), 'wb') as f:
            #     f.write(data)  # receive(conn))
            server_set(key, data)
            ok()
        elif method == 'get':
            # # if key not in database:
            # #     send(conn, b'__null__')
            # # else:
            # #     send(conn, database[key])
            # if not os.path.exists(os.path.join(save_path, key)):
            #     send(conn, b'__null__')
            # else:
            #     with open(os.path.join(save_path, key), 'rb') as f:
            #         send(conn, f.read())
            data = server_get(key)
            if len(data) < 2048:
                send(conn, server_get(key))
            else:
                send(conn, f'__big__{len(data)}'.encode())
                receive(conn)
                send(conn, data)
        elif method == 'delete':
            # # if key in database:
            # #     del database[key]
            # try:
            #     os.remove(os.path.join(save_path, key))
            # except FileNotFoundError:
            #     pass
            server_delete(key)
            ok()

    if method == 'set' or method == 'big-set':
        check_file_directory(os.path.join(save_path, key))
        with open(os.path.join(save_path, key), 'wb') as f:
            f.write(data)
        database_keys_in_order.append(key)

    while sys.getsizeof(database) > MAX_DATABASE_SIZE:
        key = database_keys_in_order.pop(0) if database_keys_in_order else next(iter(database))
        try:
            del database[key]
        except KeyError:
            pass


class Server:
    """A server for handling bucket storage requests."""
    PORT: int
    HOST: str

    def __init__(self):
        """Initialize the server with host and port from environment variables."""
        self.PORT = BUCKET_SERVER_PORT
        self.HOST = BUCKET_SERVER_HOST

    def loop(self):
        """
        Start the server loop, listening for and handling client connections.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((self.HOST, self.PORT))
            s.listen(10)
            while True:
                try:
                    conn, addr = s.accept()
                    conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                    conn.settimeout(60. * 60.)

                    threading.Thread(target=handle_client, args=(conn, )).start()
                except TimeoutError as e:
                    pass


def cleanup(slow=True):
    if USING_POSTGRES:
        postgres = Client()
        with postgres.db.connect() as conn:
            cur = conn.cursor()
            cur.execute(f'DELETE FROM {POSTGRES_TABLE} WHERE epoch + {60 * 60 * 24 * 2} > {time.time()}')
            conn.commit()
        return

    # if USING_REDIS:
    #     redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB)
    #     redis_client.delete(*[
    #         k for k in
    #         redis_client.scan_iter(match='*')
    #         if float(redis_client.get(k)) + 60 * 60 * 24 * 2 > time.time()
    #     ])

    for root, dirs, files in os.walk(save_path, topdown=True):
        dirs[:] = [d for d in dirs if d != PERSISTENT_PATH]
        for f in files:
            if os.path.getmtime(os.path.join(root, f)) + 60 * 60 * 24 > time.time():
                os.remove(os.path.join(root, f))
            if slow:
                time.sleep(0.01)



def main() -> None:
    """
    Run the bucket server.
    """
    server: Server = Server()
    print('running bucket server', server.HOST, '@', server.PORT)
    server.loop()





if __name__ == '__main__':
    main()




"""Module for parsing and executing a custom pipeline language.

This module provides functionality to parse and execute code written in a custom
pipeline language. It includes utilities for handling scopes, steps, pipes, actions,
and loops within the custom language.

Constants:
    INDENT (str): A string constant representing four spaces.
    PIPE_TOKENS (dict): A dictionary defining various tokens used in the pipeline language.
    LANGUAGES (dict): A dictionary mapping language names to their standardized identifiers.
"""
import sys
import inspect
import string
import sqlite3
import json
import traceback
from typing import Dict, Any, Generator

import yaml

# import buelon.core.step
from buelon.helpers import pipe_util, lazy_load_class
import buelon.helpers.pipe_util
from . import step_definition
from . import pipe
from . import action
from . import loop
from . import step

# Constants
INDENT = '    '
PIPE_TOKENS = {
    'import': ':',
    'pipe': '|',
    'func': ('(', ')'),
    'scope': '$',
    'str': ['"'],
    'num': ['#'],
    'bool': ['!'],
    'list': ['[', ']'],
    'call':  ['(', ')'],
    'kwargs': ['{', '}'],
    'block': [{'start': '`', 'end': '`'}]
}


LANGUAGES = {
    'python': step.LanguageTypes.python.value,
    'python3': step.LanguageTypes.python.value,
    'py': step.LanguageTypes.python.value,
    'pg': step.LanguageTypes.postgres.value,
    'postgres': step.LanguageTypes.postgres.value,
    'postgresql': step.LanguageTypes.postgres.value,
    'sqlite3': step.LanguageTypes.sqlite3.value,
    'sqlite': step.LanguageTypes.sqlite3.value
}


def delete_last_line():
    sys.stdout.write("\033[F")
    sys.stdout.write("\033[K")


def check_for_scope(index: int, line: str, variables: Dict) -> bool:
    """Check if the current line defines a scope and update variables accordingly.

    Args:
        index (int): The current line index.
        line (str): The current line of code.
        variables (dict): A dictionary containing pipeline variables.

    Returns:
        bool: True if a scope was found and processed, False otherwise.

    Raises:
        SyntaxError: If the scope syntax is invalid.
    """
    token = PIPE_TOKENS['scope']
    if line.strip().startswith(token):
        if line.count(token) > 1:
            raise SyntaxError(f'Line {index+1}: Invalid scope')
        scope = line.split(token)[1].strip()
        if not scope:
            raise SyntaxError(f'Line {index+1}: Invalid scope')
        variables['__scope__'] = scope
        return True
    return False



def run(code: str, lazy_steps: bool = False) -> Dict:
    """Parse and execute the given pipeline code.

    Args:
        code (str): The pipeline code to be executed.
        lazy_steps (bool, optional): Whether to use lazy loading for steps. Defaults to False.

    Returns:
        dict: A dictionary containing the parsed pipeline variables.

    Raises:
        Exception: If an error occurs during parsing or execution.
    """
    try:

        if lazy_steps:
            print('Lazy Steps')
            variables = lazy_load_class.LazyMap()
            lm = lazy_load_class.LazyMap()
            lm.disable_deletion()
            variables['__steps__'] = lm
        else:
            variables = {
                '__steps__': {}
            }
        variables['__starters__'] = []
        variables['__scope__'] = 'default'

        configurations = {
            'importing': False,
            'current import': None,
            'in loop': False,
            'in code block': False,
        }

        if code.strip().startswith('{'):
            try:
                i, j = pipe_util.extract_json(code)
                kwargs = json.loads(j)
                code = code[i:]
            except Exception as e:
                print(e)

        def remove_comments(line: str) -> str:
            """Remove comments from a line of code.

            Args:
                line (str): A line of code.

            Returns:
                str: The line with comments removed.
            """
            return line.split('#')[0]

        def set_index(_i: int) -> None:
            """Set the current line index.

            This function is used to update the line index, typically when
            processing loops or other constructs that may alter the normal
            flow of line-by-line execution.

            Args:
                _i (int): The new line index to set.
            """
            nonlocal i
            i = _i

        def read_line(index: int) -> None:
            """Process a single line of the pipeline code.

            Args:
                index (int): The index of the current line.
            """
            nonlocal i
            line = lines[index]

            # else:
            if step_definition.check_for_step_definition(i, line, variables, configurations):
                return

            if check_for_scope(i, line, variables):
                return

            if pipe.check_for_pipe(i, line, variables):
                return

            if action.check_for_actions(i, line, variables):
                return

            if loop.check_loop(i, line, lines, variables, read_line, set_index):
                return

            if line.strip():
                raise SyntaxError(f'Line {i+1}: Invalid syntax. `{line}`')

        lines = [remove_comments(line) for line in code.splitlines()]
        i = 0
        while i < len(lines):
            read_line(i)
            i += 1

        return variables

    except Exception as e:
        traceback.print_exc()
        print(f'Unexpected Error: {type(e).__name__} | {e}')
        exit()


def get_steps_from_code(code: str, lazy_steps: bool = False) -> Dict[str, Any]:
    """DEPRECATED: Use generate_steps_from_code() instead.

    Extract steps and starters from the given pipeline code.

    Args:
        code (str): The pipeline code to be processed.
        lazy_steps (bool, optional): Whether to use lazy loading for steps. Defaults to False.

    Raises:
        Exception: If an error occurs during parsing or execution.

    Returns:
        dict: A dictionary containing the extracted steps and starters.
    """
    import warnings
    warnings.warn("`buelon.core.pipe_interpreter.get_steps_from_code(code, lazy_steps)`. Use `buelon.core.pipe_interpreter.generate_steps_from_code(code)`.",
                  DeprecationWarning,
                  stacklevel=2)
    variables = run(code, lazy_steps=lazy_steps)
    return {'steps': variables['__steps__'], 'starters': variables['__starters__'], 'variables': variables}


allowed_chars = string.ascii_letters + string.digits + '_- '


# scope = 'default'
# priority = 0


class BuelonSyntaxError(Exception):
    pass


class BuelonBuildError(Exception):
    pass


def name_warning(name: str):
    if '  ' in name:
        print(f'Warning: `{name}` contains consecutive spaces.')


def indexes(line: str, char: str):
    r = '^' if char != '^' else '%'
    indexes = []
    while char in line:
        index = line.index(char)
        indexes.append(index)
        line = line.replace(char, '^', 1)
    return indexes


def valid_word(word: str) -> bool:
    if not word:
        return False
    return all(char in allowed_chars for char in word)


def line_is_execution(line: str):
    if (not line.startswith(' ') and line.count('(') == 1 and line.count(')') == 1 and line.index('(') < line.index(')')
            and not line.strip().startswith('for ')):
        if '=' in line and line.count('=') == 1:
            _, end = line.split('=')
            begin, end = end.split('(')
            if not begin.strip():
                return False
        return True
    return False


# steps, pipes = create_pipeline_objects(content)

# print('steps', steps)
# print('pipes', pipes)


class PipelineParser:
    """
    TODO: check for variable passing and assignment on parse
    """
    scope: str = 'default'
    priority: int = 0
    conn = None
    tab = '    '
    _build_index = 0

    def __init__(self, scope: str = 'default', priority: int = 0):
        self.scope = scope
        self.priority = priority

    def build(self, prepared_content: str):
        self.conn = sqlite3.connect('.bue/test.db')  # (':memory:')
        self.conn.execute('PRAGMA journal_mode=WAL')
        self.conn.execute('PRAGMA synchronous=NORMAL')
        self.conn.execute('DROP table if exists jobs;')
        self.conn.execute('CREATE TABLE IF NOT EXISTS jobs (id TEXT PRIMARY KEY, starter, value)')
        self.conn.execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_jobs_id ON jobs (id);')
        self.conn.commit()
        self._build_index = 0
        print(f'{self._build_index:,} compiles')
        i = 0
        # for job in self._build(yaml.load(prepared_content, yaml.Loader)):
        #     i += 1
        #     if job.parents:
        #         yield job
        #     else:
        #         self.conn.execute("INSERT INTO jobs (id, starter, value) VALUES (?, ?, ?)",
        #                           (job.id, not job.parents, json.dumps(job.to_json())))
        self._build(yaml.load(prepared_content, yaml.Loader))

        # print('deploying initiators')
        print(f"Built {self.conn.execute('SELECT count(*) FROM jobs').fetchall()[0][0]:,} jobs")
        q = 'SELECT starter, value FROM jobs order by starter, id limit {limit} offset {offset}'
        limit = 500
        offset = 0
        chunk = self.conn.execute(q.format(limit=limit, offset=offset)).fetchall()
        while chunk:
            for starter, job_json in chunk:
                i += 1
                yield step.Step().from_json(json.loads(job_json))
            offset += limit
            chunk = self.conn.execute(q.format(limit=limit, offset=offset)).fetchall()

        # print(f'{i:,} jobs')
        # for _id, s, j in self.conn.execute('SELECT * FROM jobs where starter').fetchall():
        #     print(step.Step().from_json(json.loads(j)).name)
        # print(f"Saved {self.conn.execute('SELECT count(*) FROM jobs').fetchall()[0][0]:,} jobs")
        # print(self.conn.execute('SELECT starter, count(*) FROM jobs group by starter').fetchall())
        # t = time.time()
        # q = 'SELECT starter, value FROM jobs order by starter, id limit {limit} offset {offset}'
        # limit = 1000
        # offset = 0
        # chunk = self.conn.execute(q.format(limit=limit, offset=offset)).fetchall()
        # while chunk:
        #     for starter, job_json in chunk:
        #         # yield starter, step.Step().from_json(json.loads(job_json))
        #         _ = step.Step().from_json(json.loads(job_json))
        #         del _
        #     offset += limit
        #     chunk = self.conn.execute(q.format(limit=limit, offset=offset)).fetchall()
        # print(f'{time.time() - t:0.2f} sec(s)')
        # yield '', ''

    def _build(self, variables: dict, values: dict | None = None):
        self._build_index += 1
        delete_last_line()
        print(f'{self._build_index:,} compiles')
        # varibales = {
        #     'scope': self.scope,
        #     'priority': self.priority,
        #     'job_definitions': job_definitions,
        #     'pipes': pipes,
        #     'executions': executions,
        #     'loops': loops,
        # }
        # {#pipes
        #     'name': name,
        #     'jobs': jobs
        # }
        # {#executions
        #     'result': result,
        #     'name': name,
        #     'args': args
        # }
        # {#loops
        #     'name': var,
        #     'pipe_execution': pipe_exection,
        #     'varibales': varibales,
        # }
        # {#job_definitions
        #     'name': name,
        #     'language': LANGUAGES[language],
        #     'relation': relation,
        #     'code': code,
        #     'local': local,
        #     'scope': scope,
        #     'priority': priority,
        # }
        values = values or {}

        for execution in variables['executions']:
            # for job in self.build_execution(execution, values, variables):
            #     yield job
            self.build_execution(execution, values, variables)
        for loop in variables['loops']:
            # for job in self.build_loop(loop, values, variables):
            #     yield job
            self.build_loop(loop, values, variables)

    def merge_variables(self, a: dict, b: dict) -> dict:
        # varibales = {
        #     'scope': self.scope,
        #     'priority': self.priority,
        #     'job_definitions': job_definitions,
        #     'pipes': pipes,
        #     'executions': executions,
        #     'loops': loops,
        # }
        merged = {}
        merged['scope'] = a['scope']
        merged['priority'] = a['priority']
        merged['job_definitions'] = a['job_definitions'] + b['job_definitions']
        merged['pipes'] = a['pipes'] + b['pipes']
        merged['executions'] = a['executions'] + b['executions']
        merged['loops'] = a['loops'] + b['loops']
        return merged

    def build_loop(self, loop, values, variables):
        # {#loops
        #     'name': var,
        #     'pipe_execution': pipe_exection,
        #     'varibales': varibales,
        # }
        # {#job_definitions
        #     'name': name,
        #     'language': LANGUAGES[language],
        #     'relation': relation,
        #     'code': code,
        #     'local': local,
        #     'scope': scope,
        #     'priority': priority,
        # }
        # {#pipes
        #     'name': name,
        #     'jobs': jobs
        # }
        # {#executions
        #     'result': result,
        #     'name': name,
        #     'args': args
        # }
        local = True

        if local:
            # jobs = next(self.build_execution(loop['pipe_execution'], values, variables, store_result=False))
            jobs = self.build_execution(loop['pipe_execution'], values, variables, store_result=False)

            if jobs[0].parents:
                # TODO: fix this. This only works for linear job pipes and not a pipe with arguments
                raise BuelonBuildError(f'Loop `{loop["name"]}`. Currently only single pipe loops with no arguments are supported. Example: only --> `\npipe = job1 | job2 | job3\n\nfor v in pipe():\n{self.tab}any_pipe(v)`')

            data = []

            for job in jobs:
                new_data = job.run(data).data
                del data
                del job
                data = new_data

            if not isinstance(data, list) and not inspect.isgenerator(data):
                raise BuelonBuildError(f'Loop `{loop["name"]}` must return a list or a generator. Returned `{type(data)}`')

            i = 0
            for value in data:
                job_def = self.job_for_loop(i, value, loop, variables)
                pipe = {
                    'name': loop["name"],
                    'jobs': [job_def['name']]
                }
                execution = {
                    'result': loop['name'],
                    'name': loop["name"],
                    'args': []
                }
                merged = {
                    'scope': job_def['scope'],
                    'priority': job_def['priority'],
                    'job_definitions': [job_def],
                    'pipes': [pipe],
                    'executions': [execution],
                    'loops': []
                }
                merged = self.merge_variables(merged, loop['varibales'])
                v = json.loads(json.dumps(variables))
                v['loops'] = []
                v['executions'] = []
                merged = self.merge_variables(merged, v)
                # for job in self._build(merged, values=values):
                #     yield job
                self._build(merged, values=values)
                i += 1

    def job_for_loop(self, i, value, loop, variables):
        # self.job_from_definition
        _pipe = self.get_pipe(loop['pipe_execution']['name'], variables)
        loop_job_def = self.get_job_definition(_pipe['jobs'][-1], variables)
        priority: int = loop_job_def['priority']
        scope: str = loop_job_def['scope']
        return {
            'name': f'{loop["name"]}_{i}',
            'language': 'PYTHON',
            'local': False,
            'code': f'import json\ndef main(*args):\n    return json.loads({json.dumps(json.dumps(value))})',
            'relation': 'main',
            'priority': priority,
            'scope': scope,
            'args': [],
            'result': None
        }

    def build_execution(self, execution, values, variables, append=None, store_result=True):
        for arg in execution['args']:
            if arg not in values:
                raise BuelonBuildError(
                    f'Argument `{arg}` not found. arguments can only by execution asignments. Example: `arg = pipe()`')

        jobs = []
        pipe = self.get_pipe(execution['name'], variables)
        # parents = [values[arg] for arg in execution['args']]
        parents = []

        first_job_id = buelon.helpers.pipe_util.get_id()

        for arg in execution['args']:
            job_id = values[arg]
            parents.append(job_id)
            job_json = self.conn.execute(
                'select value from jobs where id = ?', (job_id,)
            ).fetchall()[0][0]
            job = step.Step().from_json(json.loads(job_json))
            job.children.append(first_job_id)
            self.conn.execute("UPDATE jobs SET value = ? WHERE id = ?", (json.dumps(job.to_json()), job.id))
            self.conn.commit()

        last_job = None

        for job_name in pipe['jobs']:
            job_def = self.get_job_definition(job_name, variables)
            job = self.job_from_definition(job_def, provide_id=first_job_id)
            first_job_id = None
            jobs.append(job)
            job.parents = parents
            job.children = []
            if last_job:
                last_job.children.append(job.id)
            last_job = job
            parents = [job.id]

        # if append:
        #     jobs.append(job)
        #     job.parents = parents
        #     job.children = []
        #     if last_job:
        #         last_job.children.append(job.id)
        #     last_job = job

        if last_job and execution['result']:
            values[execution['result']] = last_job.id

        if not store_result:
            return jobs
            # yield jobs
        else:
            for job in jobs:
                # yield job
                self.conn.execute("INSERT INTO jobs (id, starter, value) VALUES (?, ?, ?)", (job.id, not job.parents, json.dumps(job.to_json())))
            self.conn.commit()

    def get_pipe(self, name: str, variables: dict):
        for pipe in variables['pipes']:
            if pipe['name'] == name:
                return pipe
        raise BuelonBuildError(f'Pipe `{name}` not found.')

    def get_job_definition(self, name: str, variables: dict):
        for job in variables['job_definitions']:
            if job['name'] == name:
                return job
        raise BuelonBuildError(f'Job `{name}` not found.')

    def job_from_definition(self, definition: dict, env=None, provide_id: str | None = None):
        job = step.Step()
        job.id = provide_id if provide_id else buelon.helpers.pipe_util.get_id()
        job.name = definition['name']
        job.type = definition['language']
        job.local = definition['local']
        job.code = definition['code']
        job.func = definition['relation']
        job.kwargs = env or {}
        job.priority = definition['priority']
        # job.velocity = self.velocity
        # job.tag = self.tag
        job.scope = definition['scope']
        return job

    def prepare(self, content: str):
        return yaml.dump(self.parse(content))

    def get_line_at_index(self, index: int, lines: list[str]):
        if 0 > index >= len(lines):
            raise BuelonSyntaxError(f'Error trying to read line: {index + 1}')
        return lines[index]

    def set_line_at_index(self, index: int, line: str, lines: list[str]):
        if 0 > index >= len(lines):
            raise BuelonSyntaxError(f'Error trying to set line: {index + 1}')
        lines[index] = line

    def parse(self, content: str):
        self.conn = sqlite3.connect(':memory:')
        self.conn.execute("CREATE TABLE IF NOT EXISTS in_string (start_l, start_c, end_l, end_c);")
        self.conn.commit()

        lines = content.splitlines()
        self.fetch_in_strings(lines)
        lines = self.remove_comments(lines)
        return self.get_variables(lines)

    def get_variables(self, lines: list[str]):
        self.check_for_tab_assignment(lines)

        self.check_for_scope(lines)
        self.check_for_priority(lines)

        job_definitions = self.fetch_job_definitions(lines)
        pipes = self.fetch_pipes(lines)
        executions = self.fetch_executions(lines)
        loops = self.fetch_loops(lines)
        varibales = {
            'scope': self.scope,
            'priority': self.priority,
            'job_definitions': job_definitions,
            'pipes': pipes,
            'executions': executions,
            'loops': loops,
        }

        unidentified = [f'Line {i + 1}: `{line}`' for i, line in enumerate(lines) if line.strip()]

        if unidentified:
            raise BuelonSyntaxError(' - Unidentifieable lines - \n' + '\n'.join(unidentified))

        return varibales

    def check_for_tab_assignment(self, lines: list[str]):
        for i, line in enumerate(lines):
            if line.startswith('TAB') and line.count('=') == 1:
                if line.count('"') == 2 and line.index('=') < line.index('"'):
                    self.tab = line.split('"')[1]
                    # lines[i] = ''
                    self.set_line_at_index(i, '', lines)
                if line.count("'") == 2 and line.index('=') < line.index("'"):
                    self.tab = line.split("'")[1]
                    # lines[i] = ''
                    self.set_line_at_index(i, '', lines)

    def check_for_scope(self, lines: list[str]):
        for i, line in enumerate(lines):
            if line.startswith('$'):
                self.scope = line[1:].strip()
                if not valid_word(self.scope):
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid scope name. `{line}`')
                name_warning(self.scope)
                # lines[i] = ''
                self.set_line_at_index(i, '', lines)

    def check_for_priority(self, lines: list[str]):
        for i, line in enumerate(lines):
            if line.startswith('!'):
                try:
                    self.priority = int(line[1:].strip())
                except ValueError:
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid priority value. `{line}`')
                # lines[i] = ''
                self.set_line_at_index(i, '', lines)

    def fetch_in_strings(self, lines: str):
        in_strings = []
        in_string = False
        last = None

        for i, line in enumerate(lines):
            for i1 in indexes(line, '`'):
                if in_string:
                    self.conn.execute("INSERT INTO in_string (start_l, start_c, end_l, end_c) VALUES (?, ?, ?, ?);",
                                      (*last, i, i1))
                    self.conn.commit()
                    in_string = False
                else:
                    if '#' in line and line.index('#') < i1:
                        break
                    in_string = True
                    last = (i, i1)

        return in_strings

    def in_string(self, line_index: int, char_index: int):
        cursor = self.conn.cursor()
        query = f'''
        SELECT count(*) FROM in_string 
        WHERE 
        case when start_l = {line_index} then start_c <= {char_index}
        else
            case when end_l = {line_index} then end_c >= {char_index}
            else
            start_l < {line_index}
            AND end_l > {line_index}
            end
        end;
        '''
        cursor.execute(query)
        r = cursor.fetchall()[0][0] > 0
        cursor.close()
        return r

    def remove_comments(self, lines: list[str]):
        return [
            (line.split('#')[0] if not self.in_string(i, line.index('#')) else line) if '#' in line else line
            for i, line in enumerate(lines)

        ]

    def fetch_job_definitions(self, lines: list[str]):
        definitions = []
        consumed_lines = []

        definition_starts = [
            index for index, line in enumerate(lines)
            if not line.strip().startswith('for')
               and line.strip().endswith(':')
               and not self.in_string(index, 0)
               and not line.startswith(' ')
        ]

        def check_indentation(line: str, index: int):
            if len(line) <= len(self.tab) or not line.startswith(self.tab) or line[len(self.tab)] == ' ':
                raise BuelonSyntaxError(f'Line: {index + 1} Invalid indentation. `{line}`')

        for start in definition_starts:
            scope = self.scope
            priority = self.priority

            def check_scope(i: int):
                nonlocal scope, start, consumed_lines
                check_indentation(lines[i], i)
                # line = lines[i][len(self.tab):]
                line = self.get_line_at_index(i, lines)[len(self.tab):]
                if line.startswith('$'):
                    scope = line[1:].strip()
                    if not valid_word(scope):
                        raise BuelonSyntaxError(
                            f'Line: {start + 1} Invalid char in scope for Job Definition. `{scope}`')
                    name_warning(scope)
                    consumed_lines.append(i)
                    start += 1
                    check_priority(i + 1)

            def check_priority(i: int):
                nonlocal priority, start, consumed_lines
                # check_indentation(lines[i], i)
                check_indentation(self.get_line_at_index(i, lines), i)
                # line = lines[i][len(self.tab):]
                line = self.get_line_at_index(i, lines)[len(self.tab):]
                if line.startswith('!'):
                    priority = int(line[1:].strip())
                    try:
                        int(priority)
                    except ValueError:
                        raise BuelonSyntaxError(f'Line: {start + 1} Invalid priority for Job Definition. `{priority}`')
                    consumed_lines.append(i)
                    start += 1
                    check_scope(i + 1)

            # if lines[start].count(':') > 1:
            if self.get_line_at_index(start, lines).count(':') > 1:
                raise BuelonSyntaxError(f'Line: {start + 1} Invalid name for Job Definition. More than 1 `:` found.')

            # name = lines[start].split(':')[0].strip()
            name = self.get_line_at_index(start, lines).split(':')[0].strip()
            consumed_lines.append(start)

            if not valid_word(name):
                raise BuelonSyntaxError(f'Line: {start + 1} Invalid char in name for Job Definition. `{name}`')

            check_scope(start + 1)
            check_priority(start + 1)
            # language = lines[start + 1]
            language = self.get_line_at_index(start + 1, lines)
            check_indentation(language, start + 1)
            consumed_lines.append(start + 1)

            language = language[len(self.tab):].rstrip()

            if language not in LANGUAGES:
                raise BuelonSyntaxError(
                    f'Line: {start + 2} Invalid language for Job Definition. `{language}` not found.')

            check_scope(start + 2)
            check_priority(start + 2)
            # relation = lines[start + 2]
            relation = self.get_line_at_index(start + 2, lines)
            check_indentation(relation, start + 2)
            consumed_lines.append(start + 2)

            relation = relation[len(self.tab):].rstrip()
            if not valid_word(relation):
                raise BuelonSyntaxError(f'Line: {start + 3} Invalid char in relation for Job Definition. `{relation}`')

            check_scope(start + 3)
            check_priority(start + 3)
            # check_indentation(lines[start + 3], start + 3)
            check_indentation(self.get_line_at_index(start + 3, lines), start + 3)
            consumed_lines.append(start + 3)

            local = False
            # code = lines[start + 3][len(self.tab):].rstrip()
            code = self.get_line_at_index(start + 3, lines)[len(self.tab):].rstrip()
            if not code.startswith('`'):
                local = True
                code = code.strip()
                if not code:
                    raise BuelonSyntaxError(f'Line: {start + 4} Invalid code for Job Definition. No code provided.')
            else:
                if not code.startswith('`'):
                    raise BuelonSyntaxError(
                        f'Line: {start + 4} Invalid code for Job Definition. Missing "`" at start of code.')
                elif code.count('`') > 2:
                    raise BuelonSyntaxError(
                        f'Line: {start + 4} Invalid code for Job Definition. More than 2 "`" found in "{code}"')
                elif code.count('`') == 2:
                    begin, middle, end = code.split('`')
                    if end.strip():
                        raise BuelonSyntaxError(
                            f'Line: {start + 4} Invalid code for Job Definition. Code ends with `{end.strip()}`')
                    if begin.strip():
                        raise BuelonSyntaxError(
                            f'Line: {start + 4} Invalid code for Job Definition. Code starts with `{begin.strip()}`')
                    code = middle
                else:
                    begin, end = code.split('`')
                    if begin.strip():
                        raise BuelonSyntaxError(
                            f'Line: {start + 4} Invalid code for Job Definition. Code starts with `{begin.strip()}`')
                    code = end
                    i = start + 4
                    # while i < len(lines) and '`' not in lines[i]:
                    while i < len(lines) and '`' not in self.get_line_at_index(i, lines):
                        code += '\n' + lines[i]
                        consumed_lines.append(i)
                        i += 1
                    # if lines[i].count('`') > 1:
                    if self.get_line_at_index(i, lines).count('`') > 1:
                        raise BuelonSyntaxError(
                            f'Line: {i + 1} Invalid code for Job Definition. More than 1 "`" found.')
                    consumed_lines.append(i)
                    # begin, end = lines[i].split('`')
                    begin, end = self.get_line_at_index(i, lines).split('`')
                    if end.strip():
                        raise BuelonSyntaxError(
                            f'Line: {i + 1} Invalid code for Job Definition. Code ends with `{end.strip()}`')
                    code += '\n' + begin

            name_warning(name)
            name_warning(relation)
            # print('def', {
            #     'name': name,
            #     'language': LANGUAGES[language],
            #     'relation': relation,
            #     'code': code,
            #     'local': local,
            #     'scope': scope,
            #     'priority': priority,
            # })
            definitions.append({
                'name': name,
                'language': LANGUAGES[language],
                'relation': relation,
                'code': code,
                'local': local,
                'scope': scope,
                'priority': priority,
            })

            for i in consumed_lines:
                # lines[i] = ''
                self.set_line_at_index(i, '', lines)
            consumed_lines = []
        return definitions

    def fetch_pipes(self, lines: list[str]):
        pipes = []
        consumed_lines = []

        for i, line in enumerate(lines):
            if '=' in line and not line.startswith(' ') and not line.strip().endswith(':') and not line_is_execution(
                    line):
                if line.count('=') > 1:
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe definition. More than 1 `=` found.')
                if line.split('=')[-1].strip().startswith('(') and line.count('(') == 1:

                    name, end = line.split('=')
                    name = name.strip()

                    if ')' in end:
                        end = end.strip()
                        if '|' not in end:
                            print('skip')
                            continue
                        if not end.endswith(')'):
                            raise BuelonSyntaxError(
                                f'Line: {i + 1} Invalid pipe definition. `)` missplaced in `{line}`')
                        if end.count(')') > 1:
                            raise BuelonSyntaxError(
                                f'Line: {i + 1} Invalid pipe definition. More than 1 `)` found in `{line}`')
                        jobs = end.strip()[1:-1].strip()
                        consumed_lines.append(i)
                    else:
                        code = end
                        l = [i]
                        j = i + 1
                        # nl = lines[j]
                        nl = self.get_line_at_index(j, lines)
                        l.append(j)
                        code += ' ' + nl
                        while ')' not in nl:
                            j += 1
                            # nl = lines[j]
                            nl = self.get_line_at_index(j, lines)
                            l.append(j)
                            code += ' ' + nl

                        if nl.count(')') > 1:
                            raise BuelonSyntaxError(f'Line: {j + 1} Invalid pipe definition. More than 1 `)` found.')

                        if not nl.strip().endswith(')'):
                            raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe definition. `)` missplaced in `{nl}`')

                        jobs = code.strip()[1:-1]
                        consumed_lines.extend(l)

                    if not valid_word(name):
                        raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe name. `{name}`')

                    jobs = [job.strip() for job in jobs.split('|') if job.strip() != '']
                    for job in jobs:
                        if not valid_word(job):
                            raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe job. `{job}`')
                        name_warning(job)
                    name_warning(name)
                    pipes.append({
                        'name': name,
                        'jobs': jobs
                    })

                elif '|' in line.split('=')[-1]:
                    consumed_lines.append(i)
                    name, jobs = line.split('=')
                    name = name.strip()
                    if not valid_word(name):
                        raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe name. `{name}`')
                    jobs = [job.strip() for job in jobs.split('|') if job.strip() != '']
                    for job in jobs:
                        if not valid_word(job):
                            raise BuelonSyntaxError(f'Line: {i + 1} Invalid pipe job. `{job}`')
                        name_warning(job)
                    name_warning(name)
                    pipes.append({
                        'name': name,
                        'jobs': jobs
                    })

        for i in consumed_lines:
            # lines[i] = ''
            self.set_line_at_index(i, '', lines)

        return pipes

    def fetch_executions(self, lines: list[str]):
        executions = []  # [{'name': '', args: ['', '']}...]
        consumed_lines = []

        for i, line in enumerate(lines):
            if (not line.startswith(' ') and line.count('(') == 1 and line.count(')') == 1 and line.index(
                    '(') < line.index(')')
                    and not line.strip().startswith('for ')):
                consumed_lines.append(i)
                value = line
                if '=' not in line:
                    result = None
                else:
                    if line.count('=') > 1:
                        raise BuelonSyntaxError(f'Line: {i + 1} Invalid execution definition. More than 1 `=` found.')
                    result, value = line.split('=')
                    result = result.strip()
                    if not valid_word(result):
                        raise BuelonSyntaxError(f'Line: {i + 1} Invalid execution name. `{result}`')
                name, end = value.split('(')
                args, end = end.split(')')
                if end.strip():
                    raise BuelonSyntaxError(
                        f'Line: {i + 1} Invalid execution definition. Extra chars after `)`. `{line}`')

                name = name.strip()

                if not valid_word(name):
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid execution name. `{name}`')

                args = args.split(',')
                args = [arg.strip() for arg in args if arg.strip()]
                for arg in args:
                    if not valid_word(arg):
                        raise BuelonSyntaxError(f'Line: {i + 1} Invalid execution arg. `{arg}`')
                    name_warning(arg)
                if result:
                    name_warning(result)
                name_warning(name)
                executions.append({
                    'result': result,
                    'name': name,
                    'args': args
                })

        for i in consumed_lines:
            # lines[i] = ''
            self.set_line_at_index(i, '', lines)

        return executions

    def fetch_loops(self, lines: list[str]):
        loops = []

        for i, line in enumerate(lines):
            if (line.startswith('for ')
                    and ' in ' in line and line.count(' in ') == 1
                    and line.strip().endswith(':') and line.count(':') == 1
                    and line.count('(') == 1 and line.count(')')
                    and line.index('(') < line.index(')')
                    and line.index(')') < line.index(':')
                    and line.index(' in ') < line.index('(')
                    and line.index(' in ') > line.index('for ')):

                value = line[4:].replace(':', '').strip()
                var, pipe_exection = value.split(' in ')
                var = var.strip()
                if not valid_word(var):
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid loop variable name. `{var}`')

                pipe_exection = self.fetch_executions([pipe_exection.strip()])

                if not pipe_exection:
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid loop definition. Missing pipe execution.')

                if len(pipe_exection) > 1:
                    raise BuelonSyntaxError(f'Line: {i + 1} Invalid loop definition. More than 1 pipe execution.')

                pipe_exection = pipe_exection[0]
                j = i + 1
                # nl = lines[j]
                nl = self.get_line_at_index(j, lines)
                if not nl.startswith(self.tab) or not nl.strip():
                    raise BuelonSyntaxError(f'Line: {j + 1} Invalid loop definition. Missing indentation.')

                code = nl[len(self.tab):]
                # lines[j] = ''
                self.set_line_at_index(j, '', lines)
                for j in range(j + 1, len(lines)):
                    # nl = lines[j]
                    nl = self.get_line_at_index(j, lines)
                    if nl.startswith(self.tab):
                        code += '\n' + nl[len(self.tab):]
                        # lines[j] = ''
                        self.set_line_at_index(j, '', lines)
                    else:
                        break

                # lines[i] = ''
                self.set_line_at_index(i, '', lines)
                varibales = self.get_variables([''] * (i + 1) + code.splitlines())

                loops.append({
                    'name': var,
                    'pipe_execution': pipe_exection,
                    'varibales': varibales,
                })

        return loops


def generate_steps_from_code(code: str) -> Generator[step.Step, None, None]:
    """Extract steps and starters from the given pipeline code.

    Args:
        code (str): The pipeline code to be processed.
        lazy_steps (bool, optional): Whether to use lazy loading for steps. Defaults to False.

    Raises:
        Exception: If an error occurs during parsing or execution.

    Returns:
        dict: A dictionary containing the extracted steps and starters.
    """
    # print('start')

    # with open('amazon_report_v2.bue') as f:  # ('test.bue') as f:  # ('amazon_report_v2.bue') as f:
    #     content = f.read()

    # with open('test.yaml', 'w') as f:
    #     f.write(PipelineParser().prepare(code))

    # with open('test.yaml', 'r') as f:
    #     content = f.read()
    #     for j in PipelineParser().build(content):
    #         # print(j)
    #         del j

    for job in PipelineParser().build(PipelineParser().prepare(code)):
        # print(j)
        yield job
        del job